package {

	import net.flashpunk.FP;
	import net.flashpunk.graphics.Tilemap;
	import net.flashpunk.masks.Grid;
	import flash.utils.ByteArray;

	/** More generic Tiled level parser */
	public class Tiled {

		private var twidth:int, theight:int;
		private var lwidth:int, lheight:int;
		private var tilesets:Object;	// Tileset data: name->{offset,count}
		private var layers: Object;		// Tile data
		private var objects: Object;	// Object data by layer
		private var properties: Object; // Map properties
		
		public function Tiled(data:XML=null) {
			if(data) loadLevel(data);
		}

		public function get width(): int { return lwidth; }
		public function get height(): int { return lheight; }
		public function get tileWidth(): int { return twidth; }
		public function get tileHeight(): int { return theight; }
		public function get loaded(): Boolean { return width>0 && height>0; }
		
		/** Get a tile property */
		public function property(name:String): String { return properties[name]; }


		/** Get a tile index
		 * @param name	Tile layer name
		 * @param x		x coordinate of tile
		 * @param y		y coordinate of tile
		 * @param tileset Optional tileset name
		 */
		public function tile(name:String, x:int, y:int, tileset:String=null): int {
			if(!layers[name]) return -1;
			var t:Object = { offset:1, count:512 };
			if(tileset && tilesets[tileset]) t = tilesets[tileset];
			var tile:int = layers[name][x+y*lwidth]-t.offset;
			return tile<t.count? tile: -1;
		}



		/** Create a tilemap from a layer and tileset
		 * @param name The layer name
		 * @param tiles The source image for the tiles
		 * @param tileset The name of the tileset
		 */
		public function createTilemap(name:String, tiles:*, tileset:String=''): Tilemap {
			if(!layers[name]) return null;
			var map:Tilemap = new Tilemap(tiles, lwidth*twidth, lheight*theight, twidth, theight);
			var a:Array = layers[name];
			var offset:int=1, count:int=64, used:int=0;
			if(tilesets[tileset]) {
				offset = tilesets[tileset].offset;
				count = tilesets[tileset].count;
			} else for each(var ts:Object in tilesets) {
				offset = ts.offset;
				count = ts.count;
				break;
			}
			trace('Offset:', offset, 'Count:', count);
			for(var i:int=0; i<a.length; ++i) {
				var tile:int = a[i] - offset;
				if(tile>=0 && tile<count) {
					map.setTile( i%lwidth, Math.floor(i/lwidth), tile);
					used++;
				}
			}
			return used? map: null;
		}

		/** Create a clipmap grid from a layer and tile index
		 * @param name The layer name
		 * @param tileset The name of the tileset to use
		 * @param type The index of the tile in the tileset to be the 'on' state
		 */
		public function createClipmap(name:String, tileset:String, solid:*): Grid {
			if(!layers[name]) return null;
			var grid:Grid = new Grid(lwidth*twidth, lheight*theight, twidth, theight);
			var a:Array = layers[name];
			var offset:int = tilesets[tileset]? tilesets[tileset].offset: 1;
			var tiles: int=0, s:Boolean;
			for(var i:int=0; i<a.length; ++i) {
				s = solid is Array? solid.indexOf(a[i]-offset)>=0: a[i]-offset==solid;
				grid.setTile( i%lwidth, Math.floor(i/lwidth), s);
				if(s) ++tiles;
			}
			return tiles? grid: null; //Null if no collision
		}

		/** Get an array of objects */
		public function getObjects(type:String=null, group:String=null): Array {
			var list:Array = [];
			if(group!=null) appendObjects(list, objects[group], type);
			else for each(var g:Array in objects) appendObjects(list, g, type);
			return list;
		}
		public function appendObjects(out:Array, list:Array, type:String=null): Array {
			for each(var o:Object in list) if(type==null || type==o.type) out.push(o); 
			return out;
		}



		//// //// LOADING //// ////

		private function loadLevel(data:XML): Boolean {
			FP.timeFlag();
			try {
				twidth  = data.@tilewidth;
				theight = data.@tileheight;
				lwidth  = data.@width;
				lheight = data.@height;

				//Properties
				properties = {};
				for each(var mp:XML in data.properties.property) {
					properties[mp.@name] = mp.@value;
				}


				//Tilesets
				tilesets = {};
				for each(var ts:XML in data.tileset) {
					var first:int = ts.@firstgid;
					var count:int = (ts.image.@width/ts.@tilewidth) * (ts.image.@height/ts.@tileheight);
					tilesets [ ts.@name ] = { offset: first, count: count };
				}

				// Layers
				layers = {};
				for each(var layer:XML in data.layer) {
					var layerData:Array = loadLayer(layer);
					if(layerData) layers[ layer.@name ] = layerData;
				}

				// Objects
				objects = {};
				for each(var group:XML in data.objectgroup) {
					var list:Array = objects[group.@name] = [];
					for each(var obj:XML in group.object) {
						var object:Object = { type:obj.@type, name:obj.@name, x:Number(obj.@x), y:Number(obj.@y) };
						list.push(object);
					}
				}
			}
			catch(e: Error) {
				trace(e);
				return false;
			}
			var time:uint = FP.timeFlag();
			return true;
		}

		private function loadLayer(layer:XML): Array {
			var w:int = layer.@width;
			var h:int = layer.@height;
			var tiles:Array;
			//Encoding type
			var data:String = layer.data[0];
			var encoding:String = layer.data.@encoding; //For some reason, switching directly fails
			var compression:String = layer.data.@compression;
			switch(encoding) {
			case 'base64':
				//Compression?
				switch(compression) {
				case 'none':
					tiles = makeArray( decode64(data) );
					break;
				case 'zlib':
					var tmp:ByteArray = decode64(data);
					tmp.uncompress();
					tiles = makeArray(tmp);
					break;
				default:
					throw new Error("Cannot decode compression: '" + layer.data.@compression + "'");
					break;
				}
				break;
			case 'csv':
				tiles = data.split(',');
				break;
			default:
				throw new Error("Cannot decode: '" + layer.data.@encoding + "'");
			}
			return tiles;
		}

		private static const BASE64:String = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
		private function decode64(data:String): ByteArray {
			// Create lookup array
			var lookup:Array = [];
			for(var c:int=0; c<BASE64.length; c++) lookup[BASE64.charCodeAt(c)]=c;
			// Read data: 4 characters => 1 byte
			var buffer:ByteArray = new ByteArray();
			for(var i:int=0; i<data.length-3; i+=4) {
				var a0:int = lookup[ data.charCodeAt(i+0) ];
				var a1:int = lookup[ data.charCodeAt(i+1) ];
				var a2:int = lookup[ data.charCodeAt(i+2) ];
				var a3:int = lookup[ data.charCodeAt(i+3) ];

				buffer.writeByte( (a0 << 2) | ((a1&0x30) >> 4) );
				buffer.writeByte(((a1&0x0f) << 4) | ((a2&0x3c) >> 2));
				buffer.writeByte(((a2&0x03) << 6) | a3);
			}
			buffer.position = 0;
			return buffer;
		}
		private function makeArray(raw:ByteArray): Array {
			raw.endian = 'littleEndian';
			var out:Array = [];
			while(raw.position < raw.length) out.push( raw.readInt() );
			return out;
		}
		

	}

}

